@@ -12,22 +12,53 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import os, pickle, re
+from collections import OrderedDict
+from functools import lru_cache
+import enum
+import json
+import os
+import pickle
+import re
+import shlex
+import subprocess
 import textwrap
+import typing as T
+
 from .. import build
 from .. import dependencies
 from .. import mesonlib
 from .. import mlog
-import json
-import subprocess
-from ..mesonlib import MachineChoice, MesonException, OrderedSet, OptionOverrideProxy
-from ..mesonlib import classify_unity_sources
-from ..mesonlib import File
-from ..compilers import CompilerArgs, VisualStudioLikeCompiler
-from collections import OrderedDict
-import shlex
-from functools import lru_cache
-import typing as T
+from ..mesonlib import (
+    File, MachineChoice, MesonException, OrderedSet, OptionOverrideProxy,
+    classify_unity_sources, unholder
+)
+
+if T.TYPE_CHECKING:
+    from ..interpreter import Interpreter
+
+
+class TestProtocol(enum.Enum):
+
+    EXITCODE = 0
+    TAP = 1
+    GTEST = 2
+
+    @classmethod
+    def from_str(cls, string: str) -> 'TestProtocol':
+        if string == 'exitcode':
+            return cls.EXITCODE
+        elif string == 'tap':
+            return cls.TAP
+        elif string == 'gtest':
+            return cls.GTEST
+        raise MesonException('unknown test format {}'.format(string))
+
+    def __str__(self) -> str:
+        if self is self.EXITCODE:
+            return 'exitcode'
+        elif self is self.GTEST:
+            return 'gtest'
+        return 'tap'
 
 
 class CleanTrees:
@@ -58,12 +89,13 @@ class InstallData:
         self.mesonintrospect = mesonintrospect
 
 class TargetInstallData:
-    def __init__(self, fname, outdir, aliases, strip, install_name_mappings, install_rpath, install_mode, optional=False):
+    def __init__(self, fname, outdir, aliases, strip, install_name_mappings, rpath_dirs_to_remove, install_rpath, install_mode, optional=False):
         self.fname = fname
         self.outdir = outdir
         self.aliases = aliases
         self.strip = strip
         self.install_name_mappings = install_name_mappings
+        self.rpath_dirs_to_remove = rpath_dirs_to_remove
         self.install_rpath = install_rpath
         self.install_mode = install_mode
         self.optional = optional
@@ -82,11 +114,12 @@ class ExecutableSerialisation:
 
 class TestSerialisation:
     def __init__(self, name: str, project: str, suite: str, fname: T.List[str],
-                 is_cross_built: bool, exe_wrapper: T.Optional[build.Executable],
+                 is_cross_built: bool, exe_wrapper: T.Optional[dependencies.ExternalProgram],
                  needs_exe_wrapper: bool, is_parallel: bool, cmd_args: T.List[str],
                  env: build.EnvironmentVariables, should_fail: bool,
                  timeout: T.Optional[int], workdir: T.Optional[str],
-                 extra_paths: T.List[str], protocol: str, priority: int):
+                 extra_paths: T.List[str], protocol: TestProtocol, priority: int,
+                 cmd_is_built: bool):
         self.name = name
         self.project_name = project
         self.suite = suite
@@ -105,41 +138,44 @@ class TestSerialisation:
         self.protocol = protocol
         self.priority = priority
         self.needs_exe_wrapper = needs_exe_wrapper
+        self.cmd_is_built = cmd_is_built
 
-def get_backend_from_name(backend, build):
+
+def get_backend_from_name(backend: str, build: T.Optional[build.Build] = None, interpreter: T.Optional['Interpreter'] = None) -> T.Optional['Backend']:
     if backend == 'ninja':
         from . import ninjabackend
-        return ninjabackend.NinjaBackend(build)
+        return ninjabackend.NinjaBackend(build, interpreter)
     elif backend == 'vs':
         from . import vs2010backend
-        return vs2010backend.autodetect_vs_version(build)
+        return vs2010backend.autodetect_vs_version(build, interpreter)
     elif backend == 'vs2010':
         from . import vs2010backend
-        return vs2010backend.Vs2010Backend(build)
+        return vs2010backend.Vs2010Backend(build, interpreter)
     elif backend == 'vs2015':
         from . import vs2015backend
-        return vs2015backend.Vs2015Backend(build)
+        return vs2015backend.Vs2015Backend(build, interpreter)
     elif backend == 'vs2017':
         from . import vs2017backend
-        return vs2017backend.Vs2017Backend(build)
+        return vs2017backend.Vs2017Backend(build, interpreter)
     elif backend == 'vs2019':
         from . import vs2019backend
-        return vs2019backend.Vs2019Backend(build)
+        return vs2019backend.Vs2019Backend(build, interpreter)
     elif backend == 'xcode':
         from . import xcodebackend
-        return xcodebackend.XCodeBackend(build)
+        return xcodebackend.XCodeBackend(build, interpreter)
     return None
 
 # This class contains the basic functionality that is needed by all backends.
 # Feel free to move stuff in and out of it as you see fit.
 class Backend:
-    def __init__(self, build):
+    def __init__(self, build: T.Optional[build.Build], interpreter: T.Optional['Interpreter']):
         # Make it possible to construct a dummy backend
         # This is used for introspection without a build directory
         if build is None:
             self.environment = None
             return
         self.build = build
+        self.interpreter = interpreter
         self.environment = build.environment
         self.processed_targets = {}
         self.build_dir = self.environment.get_build_dir()
@@ -163,25 +199,24 @@ class Backend:
     def get_target_filename_abs(self, target):
         return os.path.join(self.environment.get_build_dir(), self.get_target_filename(target))
 
-    def get_builtin_options_for_target(self, target):
-        return OptionOverrideProxy(target.option_overrides,
-                                   self.environment.coredata.builtins)
-
     def get_base_options_for_target(self, target):
-        return OptionOverrideProxy(target.option_overrides,
+        return OptionOverrideProxy(target.option_overrides_base,
                                    self.environment.coredata.builtins,
                                    self.environment.coredata.base_options)
 
     def get_compiler_options_for_target(self, target):
-        return OptionOverrideProxy(
-            target.option_overrides,
-            self.environment.coredata.compiler_options[target.for_machine])
+        comp_reg = self.environment.coredata.compiler_options[target.for_machine]
+        comp_override = target.option_overrides_compiler
+        return {
+            lang: OptionOverrideProxy(comp_override[lang], comp_reg[lang])
+            for lang in set(comp_reg.keys()) | set(comp_override.keys())
+        }
 
     def get_option_for_target(self, option_name, target):
-        if option_name in target.option_overrides:
-            override = target.option_overrides[option_name]
+        if option_name in target.option_overrides_base:
+            override = target.option_overrides_base[option_name]
             return self.environment.coredata.validate_option_value(option_name, override)
-        return self.environment.coredata.get_builtin_option(option_name)
+        return self.environment.coredata.get_builtin_option(option_name, target.subproject)
 
     def get_target_filename_for_linking(self, target):
         # On some platforms (msvc for instance), the file that is used for
@@ -195,7 +230,7 @@ class Backend:
             return os.path.join(self.get_target_dir(target), target.get_filename())
         elif isinstance(target, (build.CustomTarget, build.CustomTargetIndex)):
             if not target.is_linkable_target():
-                raise MesonException('Tried to link against custom target "%s", which is not linkable.' % target.name)
+                raise MesonException('Tried to link against custom target "{}", which is not linkable.'.format(target.name))
             return os.path.join(self.get_target_dir(target), target.get_filename())
         elif isinstance(target, build.Executable):
             if target.import_filename:
@@ -226,7 +261,7 @@ class Backend:
         return self.build_to_src
 
     def get_target_private_dir(self, target):
-        return os.path.join(self.get_target_dir(target), target.get_id())
+        return os.path.join(self.get_target_filename(target) + '.p')
 
     def get_target_private_dir_abs(self, target):
         return os.path.join(self.environment.get_build_dir(), self.get_target_private_dir(target))
@@ -245,19 +280,20 @@ class Backend:
         # target that the GeneratedList is used in
         return os.path.join(self.get_target_private_dir(target), src)
 
-    def get_unity_source_file(self, target, suffix):
+    def get_unity_source_file(self, target, suffix, number):
         # There is a potential conflict here, but it is unlikely that
         # anyone both enables unity builds and has a file called foo-unity.cpp.
-        osrc = target.name + '-unity.' + suffix
+        osrc = '{}-unity{}.{}'.format(target.name, number, suffix)
         return mesonlib.File.from_built_file(self.get_target_private_dir(target), osrc)
 
     def generate_unity_files(self, target, unity_src):
         abs_files = []
         result = []
         compsrcs = classify_unity_sources(target.compilers.values(), unity_src)
+        unity_size = self.get_option_for_target('unity_size', target)
 
-        def init_language_file(suffix):
-            unity_src = self.get_unity_source_file(target, suffix)
+        def init_language_file(suffix, unity_file_number):
+            unity_src = self.get_unity_source_file(target, suffix, unity_file_number)
             outfileabs = unity_src.absolute_path(self.environment.get_source_dir(),
                                                  self.environment.get_build_dir())
             outfileabs_tmp = outfileabs + '.tmp'
@@ -268,11 +304,23 @@ class Backend:
             result.append(unity_src)
             return open(outfileabs_tmp, 'w')
 
-        # For each language, generate a unity source file and return the list
+        # For each language, generate unity source files and return the list
         for comp, srcs in compsrcs.items():
-            with init_language_file(comp.get_default_suffix()) as ofile:
-                for src in srcs:
-                    ofile.write('#include<%s>\n' % src)
+            files_in_current = unity_size + 1
+            unity_file_number = 0
+            ofile = None
+            for src in srcs:
+                if files_in_current >= unity_size:
+                    if ofile:
+                        ofile.close()
+                    ofile = init_language_file(comp.get_default_suffix(), unity_file_number)
+                    unity_file_number += 1
+                    files_in_current = 0
+                ofile.write('#include<{}>\n'.format(src))
+                files_in_current += 1
+            if ofile:
+                ofile.close()
+
         [mesonlib.replace_if_different(x, x + '.tmp') for x in abs_files]
         return result
 
@@ -398,6 +446,21 @@ class Backend:
                 return True
         return False
 
+    def get_external_rpath_dirs(self, target):
+        dirs = set()
+        args = []
+        # FIXME: is there a better way?
+        for lang in ['c', 'cpp']:
+            try:
+                args.extend(self.environment.coredata.get_external_link_args(target.for_machine, lang))
+            except Exception:
+                pass
+        for arg in args:
+            if arg.startswith('-Wl,-rpath='):
+                for dir in arg.replace('-Wl,-rpath=','').split(':'):
+                    dirs.add(dir)
+        return dirs
+
     def rpaths_for_bundled_shared_libraries(self, target, exclude_system=True):
         paths = []
         for dep in target.external_deps:
@@ -412,6 +475,9 @@ class Backend:
             if exclude_system and self._libdir_is_system(libdir, target.compilers, self.environment):
                 # No point in adding system paths.
                 continue
+            # Don't remove rpaths specified in LDFLAGS.
+            if libdir in self.get_external_rpath_dirs(target):
+                continue
             # Windows doesn't support rpaths, but we use this function to
             # emulate rpaths by setting PATH, so also accept DLLs here
             if os.path.splitext(libpath)[1] not in ['.dll', '.lib', '.so', '.dylib']:
@@ -431,8 +497,15 @@ class Backend:
             result = OrderedSet()
             result.add('meson-out')
         result.update(self.rpaths_for_bundled_shared_libraries(target))
+        target.rpath_dirs_to_remove.update([d.encode('utf8') for d in result])
         return tuple(result)
 
+    @staticmethod
+    def canonicalize_filename(fname):
+        for ch in ('/', '\\', ':'):
+            fname = fname.replace(ch, '_')
+        return fname
+
     def object_filename_from_source(self, target, source):
         assert isinstance(source, mesonlib.File)
         build_dir = self.environment.get_build_dir()
@@ -457,13 +530,13 @@ class Backend:
             source = 'meson-generated_' + os.path.relpath(rel_src, targetdir)
         else:
             if os.path.isabs(rel_src):
-                # Not from the source directory; hopefully this doesn't conflict with user's source files.
-                source = os.path.basename(rel_src)
+                # Use the absolute path directly to avoid file name conflicts
+                source = rel_src
             else:
                 source = os.path.relpath(os.path.join(build_dir, rel_src),
                                          os.path.join(self.environment.get_source_dir(), target.get_subdir()))
         machine = self.environment.machines[target.for_machine]
-        return source.replace('/', '_').replace('\\', '_') + '.' + machine.get_object_suffix()
+        return self.canonicalize_filename(source) + '.' + machine.get_object_suffix()
 
     def determine_ext_objs(self, extobj, proj_dir_to_build_root):
         result = []
@@ -491,16 +564,18 @@ class Backend:
 
         targetdir = self.get_target_private_dir(extobj.target)
 
-        # With unity builds, there's just one object that contains all the
-        # sources, and we only support extracting all the objects in this mode,
-        # so just return that.
+        # With unity builds, sources don't map directly to objects,
+        # we only support extracting all the objects in this mode,
+        # so just return all object files.
         if self.is_unity(extobj.target):
             compsrcs = classify_unity_sources(extobj.target.compilers.values(), sources)
             sources = []
-            for comp in compsrcs.keys():
-                osrc = self.get_unity_source_file(extobj.target,
-                                                  comp.get_default_suffix())
-                sources.append(osrc)
+            unity_size = self.get_option_for_target('unity_size', extobj.target)
+            for comp, srcs in compsrcs.items():
+                for i in range(len(srcs) // unity_size + 1):
+                    osrc = self.get_unity_source_file(extobj.target,
+                                                      comp.get_default_suffix(), i)
+                    sources.append(osrc)
 
         for osrc in sources:
             objname = self.object_filename_from_source(extobj.target, osrc)
@@ -521,14 +596,14 @@ class Backend:
     def create_msvc_pch_implementation(self, target, lang, pch_header):
         # We have to include the language in the file name, otherwise
         # pch.c and pch.cpp will both end up as pch.obj in VS backends.
-        impl_name = 'meson_pch-%s.%s' % (lang, lang)
+        impl_name = 'meson_pch-{}.{}'.format(lang, lang)
         pch_rel_to_build = os.path.join(self.get_target_private_dir(target), impl_name)
         # Make sure to prepend the build dir, since the working directory is
         # not defined. Otherwise, we might create the file in the wrong path.
         pch_file = os.path.join(self.build_dir, pch_rel_to_build)
         os.makedirs(os.path.dirname(pch_file), exist_ok=True)
 
-        content = '#include "%s"' % os.path.basename(pch_header)
+        content = '#include "{}"'.format(os.path.basename(pch_header))
         pch_file_tmp = pch_file + '.tmp'
         with open(pch_file_tmp, 'w') as f:
             f.write(content)
@@ -537,38 +612,22 @@ class Backend:
 
     @staticmethod
     def escape_extra_args(compiler, args):
-        # No extra escaping/quoting needed when not running on Windows
-        if not mesonlib.is_windows():
-            return args
+        # all backslashes in defines are doubly-escaped
         extra_args = []
-        # Compiler-specific escaping is needed for -D args but not for any others
-        if isinstance(compiler, VisualStudioLikeCompiler):
-            # MSVC needs escaping when a -D argument ends in \ or \"
-            for arg in args:
-                if arg.startswith('-D') or arg.startswith('/D'):
-                    # Without extra escaping for these two, the next character
-                    # gets eaten
-                    if arg.endswith('\\'):
-                        arg += '\\'
-                    elif arg.endswith('\\"'):
-                        arg = arg[:-2] + '\\\\"'
-                extra_args.append(arg)
-        else:
-            # MinGW GCC needs all backslashes in defines to be doubly-escaped
-            # FIXME: Not sure about Cygwin or Clang
-            for arg in args:
-                if arg.startswith('-D') or arg.startswith('/D'):
-                    arg = arg.replace('\\', '\\\\')
-                extra_args.append(arg)
+        for arg in args:
+            if arg.startswith('-D') or arg.startswith('/D'):
+                arg = arg.replace('\\', '\\\\')
+            extra_args.append(arg)
+
         return extra_args
 
     def generate_basic_compiler_args(self, target, compiler, no_warn_args=False):
         # Create an empty commands list, and start adding arguments from
         # various sources in the order in which they must override each other
         # starting from hard-coded defaults followed by build options and so on.
-        commands = CompilerArgs(compiler)
+        commands = compiler.compiler_args()
 
-        copt_proxy = self.get_compiler_options_for_target(target)
+        copt_proxy = self.get_compiler_options_for_target(target)[compiler.language]
         # First, the trivial ones that are impossible to override.
         #
         # Add -nostdinc/-nostdinc++ if needed; can't be overridden
@@ -648,7 +707,7 @@ class Backend:
         args = []
         for d in deps:
             if not (d.is_linkable_target()):
-                raise RuntimeError('Tried to link with a non-library target "%s".' % d.get_basename())
+                raise RuntimeError('Tried to link with a non-library target "{}".'.format(d.get_basename()))
             arg = self.get_target_filename_for_linking(d)
             if not arg:
                 continue
@@ -689,6 +748,7 @@ class Backend:
             for deppath in self.rpaths_for_bundled_shared_libraries(target, exclude_system=False):
                 result.add(os.path.normpath(os.path.join(self.environment.get_build_dir(), deppath)))
         for bdep in extra_bdeps:
+            prospectives.add(bdep)
             prospectives.update(bdep.get_transitive_link_deps())
         # Internal deps
         for ld in prospectives:
@@ -721,7 +781,16 @@ class Backend:
                 # E.g. an external verifier or simulator program run on a generated executable.
                 # Can always be run without a wrapper.
                 test_for_machine = MachineChoice.BUILD
-            is_cross = not self.environment.machines.matches_build_machine(test_for_machine)
+
+            # we allow passing compiled executables to tests, which may be cross built.
+            # We need to consider these as well when considering whether the target is cross or not.
+            for a in t.cmd_args:
+                if isinstance(a, build.BuildTarget):
+                    if a.for_machine is MachineChoice.HOST:
+                        test_for_machine = MachineChoice.HOST
+                        break
+
+            is_cross = self.environment.is_cross_build(test_for_machine)
             if is_cross and self.environment.need_exe_wrapper():
                 exe_wrapper = self.environment.get_exe_wrapper()
             else:
@@ -734,10 +803,9 @@ class Backend:
                 extra_paths = self.determine_windows_extra_paths(exe, extra_bdeps)
             else:
                 extra_paths = []
+
             cmd_args = []
-            for a in t.cmd_args:
-                if hasattr(a, 'held_object'):
-                    a = a.held_object
+            for a in unholder(t.cmd_args):
                 if isinstance(a, build.BuildTarget):
                     extra_paths += self.determine_windows_extra_paths(a, [])
                 if isinstance(a, mesonlib.File):
@@ -745,6 +813,11 @@ class Backend:
                     cmd_args.append(a)
                 elif isinstance(a, str):
                     cmd_args.append(a)
+                elif isinstance(a, build.Executable):
+                    p = self.construct_target_rel_path(a, t.workdir)
+                    if p == a.get_filename():
+                        p = './' + p
+                    cmd_args.append(p)
                 elif isinstance(a, build.Target):
                     cmd_args.append(self.construct_target_rel_path(a, t.workdir))
                 else:
@@ -753,7 +826,8 @@ class Backend:
                                    exe_wrapper, self.environment.need_exe_wrapper(),
                                    t.is_parallel, cmd_args, t.env,
                                    t.should_fail, t.timeout, t.workdir,
-                                   extra_paths, t.protocol, t.priority)
+                                   extra_paths, t.protocol, t.priority,
+                                   isinstance(exe, build.Executable))
             arr.append(ts)
         return arr
 
@@ -787,17 +861,28 @@ class Backend:
             deps.extend(self.environment.coredata.cross_files)
         deps.extend(self.environment.coredata.config_files)
         deps.append('meson-private/coredata.dat')
-        if os.path.exists(os.path.join(self.environment.get_source_dir(), 'meson_options.txt')):
-            deps.append(os.path.join(self.build_to_src, 'meson_options.txt'))
-        for sp in self.build.subprojects.keys():
-            fname = os.path.join(self.environment.get_source_dir(), sp, 'meson_options.txt')
-            if os.path.isfile(fname):
-                deps.append(os.path.join(self.build_to_src, sp, 'meson_options.txt'))
+        self.check_clock_skew(deps)
         return deps
 
-    def exe_object_to_cmd_array(self, exe):
-        if isinstance(exe, build.BuildTarget):
-            if exe.for_machine is not MachineChoice.BUILD:
+    def check_clock_skew(self, file_list):
+        # If a file that leads to reconfiguration has a time
+        # stamp in the future, it will trigger an eternal reconfigure
+        # loop.
+        import time
+        now = time.time()
+        for f in file_list:
+            absf = os.path.join(self.environment.get_build_dir(), f)
+            ftime = os.path.getmtime(absf)
+            delta = ftime - now
+            # On Windows disk time stamps sometimes point
+            # to the future by a minuscule amount, less than
+            # 0.001 seconds. I don't know why.
+            if delta > 0.001:
+                raise MesonException('Clock skew detected. File {} has a time stamp {:.4f}s in the future.'.format(absf, delta))
+
+    def build_target_to_cmd_array(self, bt):
+        if isinstance(bt, build.BuildTarget):
+            if isinstance(bt, build.Executable) and bt.for_machine is not MachineChoice.BUILD:
                 if (self.environment.is_cross_build() and
                         self.environment.exe_wrapper is None and
                         self.environment.need_exe_wrapper()):
@@ -805,12 +890,12 @@ class Backend:
                         Cannot use target {} as a generator because it is built for the
                         host machine and no exe wrapper is defined or needs_exe_wrapper is
                         true. You might want to set `native: true` instead to build it for
-                        the build machine.'''.format(exe.name))
+                        the build machine.'''.format(bt.name))
                     raise MesonException(s)
-            exe_arr = [os.path.join(self.environment.get_build_dir(), self.get_target_filename(exe))]
+            arr = [os.path.join(self.environment.get_build_dir(), self.get_target_filename(bt))]
         else:
-            exe_arr = exe.get_command()
-        return exe_arr
+            arr = bt.get_command()
+        return arr
 
     def replace_extra_args(self, args, genlist):
         final_args = []
@@ -828,7 +913,7 @@ class Backend:
             m = regex.search(arg)
             while m is not None:
                 index = int(m.group(1))
-                src = '@OUTPUT%d@' % index
+                src = '@OUTPUT{}@'.format(index)
                 arg = arg.replace(src, os.path.join(private_dir, output_list[index]))
                 m = regex.search(arg)
             newargs.append(arg)
@@ -844,14 +929,10 @@ class Backend:
         # also be built by default. XXX: Sometime in the future these should be
         # built only before running tests.
         for t in self.build.get_tests():
-            exe = t.exe
-            if hasattr(exe, 'held_object'):
-                exe = exe.held_object
+            exe = unholder(t.exe)
             if isinstance(exe, (build.CustomTarget, build.BuildTarget)):
                 result[exe.get_id()] = exe
-            for arg in t.cmd_args:
-                if hasattr(arg, 'held_object'):
-                    arg = arg.held_object
+            for arg in unholder(t.cmd_args):
                 if not isinstance(arg, (build.CustomTarget, build.BuildTarget)):
                     continue
                 result[arg.get_id()] = arg
@@ -891,9 +972,7 @@ class Backend:
         Returns the path to them relative to the build root directory.
         '''
         srcs = []
-        for i in target.get_sources():
-            if hasattr(i, 'held_object'):
-                i = i.held_object
+        for i in unholder(target.get_sources()):
             if isinstance(i, str):
                 fname = [os.path.join(self.build_to_src, target.subdir, i)]
             elif isinstance(i, build.BuildTarget):
@@ -944,8 +1023,8 @@ class Backend:
         # Evaluate the command list
         cmd = []
         for i in target.command:
-            if isinstance(i, build.Executable):
-                cmd += self.exe_object_to_cmd_array(i)
+            if isinstance(i, build.BuildTarget):
+                cmd += self.build_target_to_cmd_array(i)
                 continue
             elif isinstance(i, build.CustomTarget):
                 # GIR scanner will attempt to execute this binary but
@@ -961,35 +1040,36 @@ class Backend:
             elif not isinstance(i, str):
                 err_msg = 'Argument {0} is of unknown type {1}'
                 raise RuntimeError(err_msg.format(str(i), str(type(i))))
-            elif '@SOURCE_ROOT@' in i:
-                i = i.replace('@SOURCE_ROOT@', source_root)
-            elif '@BUILD_ROOT@' in i:
-                i = i.replace('@BUILD_ROOT@', build_root)
-            elif '@DEPFILE@' in i:
-                if target.depfile is None:
-                    msg = 'Custom target {!r} has @DEPFILE@ but no depfile ' \
-                          'keyword argument.'.format(target.name)
-                    raise MesonException(msg)
-                dfilename = os.path.join(outdir, target.depfile)
-                i = i.replace('@DEPFILE@', dfilename)
-            elif '@PRIVATE_DIR@' in i:
-                if target.absolute_paths:
-                    pdir = self.get_target_private_dir_abs(target)
-                else:
-                    pdir = self.get_target_private_dir(target)
-                i = i.replace('@PRIVATE_DIR@', pdir)
-            elif '@PRIVATE_OUTDIR_' in i:
-                match = re.search(r'@PRIVATE_OUTDIR_(ABS_)?([^/\s*]*)@', i)
-                if not match:
-                    msg = 'Custom target {!r} has an invalid argument {!r}' \
-                          ''.format(target.name, i)
-                    raise MesonException(msg)
-                source = match.group(0)
-                if match.group(1) is None and not target.absolute_paths:
-                    lead_dir = ''
-                else:
-                    lead_dir = self.environment.get_build_dir()
-                i = i.replace(source, os.path.join(lead_dir, outdir))
+            else:
+                if '@SOURCE_ROOT@' in i:
+                    i = i.replace('@SOURCE_ROOT@', source_root)
+                if '@BUILD_ROOT@' in i:
+                    i = i.replace('@BUILD_ROOT@', build_root)
+                if '@DEPFILE@' in i:
+                    if target.depfile is None:
+                        msg = 'Custom target {!r} has @DEPFILE@ but no depfile ' \
+                              'keyword argument.'.format(target.name)
+                        raise MesonException(msg)
+                    dfilename = os.path.join(outdir, target.depfile)
+                    i = i.replace('@DEPFILE@', dfilename)
+                if '@PRIVATE_DIR@' in i:
+                    if target.absolute_paths:
+                        pdir = self.get_target_private_dir_abs(target)
+                    else:
+                        pdir = self.get_target_private_dir(target)
+                    i = i.replace('@PRIVATE_DIR@', pdir)
+                if '@PRIVATE_OUTDIR_' in i:
+                    match = re.search(r'@PRIVATE_OUTDIR_(ABS_)?([^/\s*]*)@', i)
+                    if not match:
+                        msg = 'Custom target {!r} has an invalid argument {!r}' \
+                              ''.format(target.name, i)
+                        raise MesonException(msg)
+                    source = match.group(0)
+                    if match.group(1) is None and not target.absolute_paths:
+                        lead_dir = ''
+                    else:
+                        lead_dir = self.environment.get_build_dir()
+                    i = i.replace(source, os.path.join(lead_dir, outdir))
             cmd.append(i)
         # Substitute the rest of the template strings
         values = mesonlib.get_filenames_templates_dict(inputs, outputs)
@@ -1028,7 +1108,7 @@ class Backend:
             subprocess.check_call(cmd, env=child_env)
 
     def create_install_data(self):
-        strip_bin = self.environment.binaries.host.lookup_entry('strip')
+        strip_bin = self.environment.lookup_binary_entry(MachineChoice.HOST, 'strip')
         if strip_bin is None:
             if self.environment.is_cross_build():
                 mlog.warning('Cross file does not specify strip binary, result will not be stripped.')
@@ -1090,6 +1170,7 @@ class Backend:
                     mappings = t.get_link_deps_mapping(d.prefix, self.environment)
                     i = TargetInstallData(self.get_target_filename(t), outdirs[0],
                                           t.get_aliases(), should_strip, mappings,
+                                          t.rpath_dirs_to_remove,
                                           t.install_rpath, install_mode)
                     d.targets.append(i)
 
@@ -1107,14 +1188,14 @@ class Backend:
                                 implib_install_dir = self.environment.get_import_lib_dir()
                             # Install the import library; may not exist for shared modules
                             i = TargetInstallData(self.get_target_filename_for_linking(t),
-                                                  implib_install_dir, {}, False, {}, '', install_mode,
+                                                  implib_install_dir, {}, False, {}, set(), '', install_mode,
                                                   optional=isinstance(t, build.SharedModule))
                             d.targets.append(i)
 
                         if not should_strip and t.get_debug_filename():
                             debug_file = os.path.join(self.get_target_dir(t), t.get_debug_filename())
                             i = TargetInstallData(debug_file, outdirs[0],
-                                                  {}, False, {}, '',
+                                                  {}, False, {}, set(), '',
                                                   install_mode, optional=True)
                             d.targets.append(i)
                 # Install secondary outputs. Only used for Vala right now.
@@ -1124,7 +1205,7 @@ class Backend:
                         if outdir is False:
                             continue
                         f = os.path.join(self.get_target_dir(t), output)
-                        i = TargetInstallData(f, outdir, {}, False, {}, None, install_mode)
+                        i = TargetInstallData(f, outdir, {}, False, {}, set(), None, install_mode)
                         d.targets.append(i)
             elif isinstance(t, build.CustomTarget):
                 # If only one install_dir is specified, assume that all
@@ -1137,7 +1218,7 @@ class Backend:
                 if num_outdirs == 1 and num_out > 1:
                     for output in t.get_outputs():
                         f = os.path.join(self.get_target_dir(t), output)
-                        i = TargetInstallData(f, outdirs[0], {}, False, {}, None, install_mode,
+                        i = TargetInstallData(f, outdirs[0], {}, False, {}, set(), None, install_mode,
                                               optional=not t.build_by_default)
                         d.targets.append(i)
                 else:
@@ -1146,7 +1227,7 @@ class Backend:
                         if outdir is False:
                             continue
                         f = os.path.join(self.get_target_dir(t), output)
-                        i = TargetInstallData(f, outdir, {}, False, {}, None, install_mode,
+                        i = TargetInstallData(f, outdir, {}, False, {}, set(), None, install_mode,
                                               optional=not t.build_by_default)
                         d.targets.append(i)
 
