@@ -426,8 +426,8 @@ class IfNode(BaseNode):
 class IfClauseNode(BaseNode):
     def __init__(self, linenode: BaseNode):
         super().__init__(linenode.lineno, linenode.colno, linenode.filename)
-        self.ifs = []  # type: T.List[IfNode]
-        self.elseblock = EmptyNode(linenode.lineno, linenode.colno, linenode.filename)  # type: T.Union[EmptyNode, CodeBlockNode]
+        self.ifs = []          # type: T.List[IfNode]
+        self.elseblock = None  # type: T.Union[EmptyNode, CodeBlockNode]
 
 class UMinusNode(BaseNode):
     def __init__(self, current_location: Token, value: BaseNode):
@@ -489,6 +489,13 @@ class Parser:
             return True
         return False
 
+    def accept_any(self, tids: T.Sequence[str]) -> str:
+        tid = self.current.tid
+        if tid in tids:
+            self.getsym()
+            return tid
+        return ''
+
     def expect(self, s: str) -> bool:
         if self.accept(s):
             return True
@@ -562,36 +569,35 @@ class Parser:
         return left
 
     def e5(self) -> BaseNode:
-        return self.e5add()
+        return self.e5addsub()
 
-    def e5add(self) -> BaseNode:
-        left = self.e5sub()
-        if self.accept('plus'):
-            return ArithmeticNode('add', left, self.e5add())
+    def e5addsub(self) -> BaseNode:
+        op_map = {
+            'plus': 'add',
+            'dash': 'sub',
+        }
+        left = self.e5muldiv()
+        while True:
+            op = self.accept_any(tuple(op_map.keys()))
+            if op:
+                left = ArithmeticNode(op_map[op], left, self.e5muldiv())
+            else:
+                break
         return left
 
-    def e5sub(self) -> BaseNode:
-        left = self.e5mod()
-        if self.accept('dash'):
-            return ArithmeticNode('sub', left, self.e5sub())
-        return left
-
-    def e5mod(self) -> BaseNode:
-        left = self.e5mul()
-        if self.accept('percent'):
-            return ArithmeticNode('mod', left, self.e5mod())
-        return left
-
-    def e5mul(self) -> BaseNode:
-        left = self.e5div()
-        if self.accept('star'):
-            return ArithmeticNode('mul', left, self.e5mul())
-        return left
-
-    def e5div(self) -> BaseNode:
+    def e5muldiv(self) -> BaseNode:
+        op_map = {
+            'percent': 'mod',
+            'star': 'mul',
+            'fslash': 'div',
+        }
         left = self.e6()
-        if self.accept('fslash'):
-            return ArithmeticNode('div', left, self.e5div())
+        while True:
+            op = self.accept_any(tuple(op_map.keys()))
+            if op:
+                left = ArithmeticNode(op_map[op], left, self.e6())
+            else:
+                break
         return left
 
     def e6(self) -> BaseNode:
@@ -741,9 +747,7 @@ class Parser:
         block = self.codeblock()
         clause.ifs.append(IfNode(clause, condition, block))
         self.elseifblock(clause)
-        elseblock = self.elseblock()
-        if elseblock:
-            clause.elseblock = elseblock
+        clause.elseblock = self.elseblock()
         return clause
 
     def elseifblock(self, clause) -> None:
@@ -753,11 +757,11 @@ class Parser:
             b = self.codeblock()
             clause.ifs.append(IfNode(s, s, b))
 
-    def elseblock(self) -> T.Optional[CodeBlockNode]:
+    def elseblock(self) -> T.Union[CodeBlockNode, EmptyNode]:
         if self.accept('else'):
             self.expect('eol')
             return self.codeblock()
-        return None
+        return EmptyNode(self.current.lineno, self.current.colno, self.current.filename)
 
     def line(self) -> BaseNode:
         block_start = self.current
